unit MyMetal;
//Metal Offscreen Rendering
{$mode objfpc}{$H+}
{$modeswitch objectivec1}
{$modeswitch advancedrecords}

{$define UseManagedRT}    //  Offscreen texture RT is managed and we copy directly RT to our buffer

interface

uses
  Metal, MetalKit, {SimdUtils,} VectorMath, LResources, Dialogs, Classes,
  SysUtils, Controls, Graphics, math, CocoaAll, CTypes, MacOSAll, shaders;

type
  { TMetalLibrary }
  TMetalLibrary = class (TObject)
  public
    lib      : MTLLibraryProtocol;
    functions: NSMutableDictionary;

    function GetFunction (name: string): MTLFunctionProtocol;
    destructor Destroy; override;
  end;

  { TMetalPipeline }
  TMetalPipeline = class (TObject)
  public
    renderPipelineState : MTLRenderPipelineStateProtocol;
    computePipelineState: MTLComputePipelineStateProtocol;
    depthStencilState   : MTLDepthStencilStateProtocol;
    shaderLibrary       : TMetalLibrary;

    destructor Destroy; override;
  end;

  { TMetalLibraryOptions }
  TMetalLibraryOptions = record
  public
    name: string;
    code: string;

    class function Default: TMetalLibraryOptions; static;
    constructor Create(_name,_code: string);
  end;

  { TMetalPipelineOptions }
  TMetalPipelineOptions = record
  public
    libraryName: string;
    code       : AnsiString;
    shaderLibrary: TMetalLibrary;

    vertexShader: string;
    fragmentShader: string;
    kernelFunction: string;
    vertexDescriptor: MTLVertexDescriptor;
    pipelineDescriptor: MTLRenderPipelineDescriptor;

    class function Default: TMetalPipelineOptions; static;
  end;

  { TMetalContext }
  TMetalContext = class (TObject)
  private
    type TFrameState = (kMetalContextFrameStateOpen, kMetalContextFrameStateRender);
  public
    OnPaint        : TNotifyEvent;

    device         : MTLDeviceProtocol;
    commandQueue   : MTLCommandQueueProtocol;
    currentPipeline: TMetalPipeline;
    textureLoader  : MTKTextureLoader;

    commandBuffer  : MTLCommandBufferProtocol;
    renderEncoder  : MTLRenderCommandEncoderProtocol;
    computeEncoder : MTLComputeCommandEncoderProtocol;
    frameState     : set of TFrameState;

    class function SharedContext: TMetalContext;

    procedure SetPreferredFrameRate(newValue: integer);
    procedure SetColorPixelFormat(pixelFormat: MTLPixelFormat);
    procedure MakeCurrent;
    procedure Draw;

    destructor Destroy; override;
  end;

  { TRTTexture }
  TRTTexture = class (TObject)
  private
    RT                : MTLTextureProtocol;
    ColorPipeline     : TMetalPipeline;
    RndPass           : MTLRenderPassDescriptor;
    w,h               : integer;

    procedure InitShader;
    procedure SaveToBmp(const FN: AnsiString);
  public

    constructor Create(Width, Height: Integer);
    procedure DrawTo;
  end;

  { Categories }
  MTLRenderPassDescriptorHelper = objccategory (MTLRenderPassDescriptor)
    function colorAttachmentAtIndex(index: integer): MTLRenderPassColorAttachmentDescriptor; message 'colorAttachmentAtIndex:';
  end;

  MTLRenderPipelineDescriptorHelper = objccategory (MTLRenderPipelineDescriptor)
    function colorAttachmentAtIndex(index: integer): MTLRenderPipelineColorAttachmentDescriptor; message 'colorAttachmentAtIndex:';
  end;


function MTLCreateContext: TMetalContext; overload; inline;

function ResourcePath(name: pchar; ofType: pchar): string;
procedure Fatal(condition: boolean; msg: string = ''; error: NSError = nil);
procedure Fatal(msg: string; error: NSError = nil);
procedure Show(obj: id; msg: string = '');

implementation

threadvar
  CurrentThreadContext: TMetalContext;

const
  kError_InvalidContext = 'no current context';

function CFStringToStr(AString: CFStringRef; Encoding: CFStringEncoding = kCFStringEncodingUTF8): WideString;
var
  l: integer;
begin
  Result:='';
  if AString=nil then Exit;
  l:=CFStringGetLength(AString);
  SetLength(Result, l);
  CFStringGetCharacters(AString, CFRangeMake(0, l), @Result[1]);
end;

function NSStringToString(ns: NSString): WideString;
begin
  Result:=CFStringToStr(CFStringRef(ns));
end;

type
 { TVertUniforms = record //Uniforms for vertex shader
    viewportSize: TVec2;
  end;}
  TAAPLVertex = record
    position: TVec4;
    // align each vertex attribute on 16 byte boundries
    //padding : TVec3;
    color   : TVec4;
  end;

function AAPLVertex(constref position: TVec3; constref color: TVec4): TAAPLVertex;
begin
  result.position:=Vec4(position, 1.0);
  result.color:=color;
end;

function NSSTR(str: string): NSString; overload;
begin
  Result:=NSString.stringWithCString_length(@str[1], length(str));
end;

function ResourceURL(name: pchar; ofType: pchar): NSURL;
var
  url: NSURL;
begin
  result:=NSBundle.mainBundle.URLForResource_withExtension(NSSTR(name), NSSTR(ofType));
end;

function ResourcePath(name: pchar; ofType: pchar): string;
var
  url: NSURL;
begin
  url:=ResourceURL(name, ofType);
  result:=url.relativePath.UTF8String;
end;

function BundleDir: string;
var
  i, l: integer;
begin
  Result:=NSBundle.mainBundle.bundlepath.UTF8String;
  l:=Length(Result);
  i:=l;
  while i>0 do begin
    if Result[i]=pathdelim then begin
      Result:=Copy(Result, 1, i);
      Exit;
    end;
    dec(i);
  end;
end;

procedure Show(obj: id; msg: string='');
begin
  if msg<>'' then msg:=msg+' ';
  if obj=nil then writeln(msg+'nil')
  else writeln(msg,obj.description.UTF8String);
end;

procedure Fatal(msg: string; error: NSError=nil);
begin
  Fatal(true, msg, error);
end;

procedure Fatal(condition: boolean; msg: string=''; error: NSError=nil);
begin
  if condition then begin
    if error=nil then writeln(msg)
    else writeln(msg+' -> '+error.localizedDescription.UTF8String);
    raise Exception.Create('fatal');
  end;
end;

function LocalCreatePipelineDescriptor: MTLRenderPipelineDescriptor;
{const
  AAPLVertexAttributePosition = 0;
  AAPLVertexAttributeTexcoord = 1;}
var
  pipelineStateDescriptor: MTLRenderPipelineDescriptor;
  //vertexDesc             : MTLVertexDescriptor;
  colorAttachment        : MTLRenderPipelineColorAttachmentDescriptor;
//  view                   : MTKView;
begin
  pipelineStateDescriptor:=MTLRenderPipelineDescriptor.alloc.init.autorelease;
 { vertexDesc:=MTLVertexDescriptor.alloc.init;

  //Positions
  vertexDesc.attributes.objectAtIndexedSubscript(AAPLVertexAttributePosition).setformat(MTLVertexFormatFloat3);
  vertexDesc.attributes.objectAtIndexedSubscript(AAPLVertexAttributePosition).setoffset(0);
  //vertexDesc.attributes.objectAtIndexedSubscript(AAPLVertexAttributePosition).bufferIndex:=AAPLBufferIndexMeshPositions;

  //Texture coordinates
  vertexDesc.attributes.objectAtIndexedSubscript(AAPLVertexAttributeTexcoord).setformat(MTLVertexFormatFloat2);
  vertexDesc.attributes.objectAtIndexedSubscript(AAPLVertexAttributeTexcoord).setoffset(0);  }
 // vertexDesc.attributes.objectAtIndexedSubscript(AAPLVertexAttributeTexcoord).bufferIndex:=AAPLBufferIndexMeshGenerics;

  pipelineStateDescriptor.setDepthAttachmentPixelFormat(MTLPixelFormatDepth32Float);
 // pipelineStateDescriptor.setvertexDescriptor(vertexDesc);
//  pipelineStateDescriptor.setSampleCount(view.sampleCount);

  colorAttachment:=pipelineStateDescriptor.colorAttachments.objectAtIndexedSubscript(0);
  colorAttachment.setPixelFormat(MTLPixelFormatBGRA8Unorm);

  result:=pipelineStateDescriptor;
end;

function MTLRenderPassDescriptorHelper.colorAttachmentAtIndex(index: integer): MTLRenderPassColorAttachmentDescriptor;
begin
  result:=self.colorAttachments.objectAtIndexedSubscript(index);
end;

function MTLRenderPipelineDescriptorHelper.colorAttachmentAtIndex(index: integer): MTLRenderPipelineColorAttachmentDescriptor;
begin
  result:=self.colorAttachments.objectAtIndexedSubscript(index);
end;

procedure LogMetalError(msg: string);
var
  err: NSError;
begin
  err:=CurrentThreadContext.commandbuffer.error;
  if err.code<>0 then writeln('MTLError= '+ WideString(IntToStr(err.code))+', '+ NSStringToString(err.domain)+', '+WideString(msg));
end;

function LocalCreateContext(device: MTLDeviceProtocol): TMetalContext;
var
  context: TMetalContext;
begin
  context:=TMetalContext.Create;
//  context.view:=view;
  context.device:=device;

  Fatal(context.device = nil, 'no gpu device found.');
  Show(context.device, 'GPU:');

  context.commandQueue:=context.device.newCommandQueue;

  {view.setColorPixelFormat(MTLPixelFormatBGRA8Unorm);
  view.setDepthStencilPixelFormat(MTLPixelFormatDepth32Float);}

  result:=context;
end;

function MTLCreateContext({control: TMetalControl}): TMetalContext;
begin
  result:=LocalCreateContext(MTLCreateSystemDefaultDevice);
  Result.MakeCurrent;
end;

function LocalCreateLibrary(options: TMetalLibraryOptions): TMetalLibrary;

  function CompileShader (device: MTLDeviceProtocol; name: string): MTLLibraryProtocol;
  var
    source: NSString;
    error : NSError;
  begin
    source:=NSString.stringWithContentsOfFile_encoding_error(NSSTR(name), NSUTF8StringEncoding, @error);
    Fatal(source = nil, 'error loading library file', error);
    result:=device.newLibraryWithSource_options_error(source, nil, @error);
    Fatal(result = nil, 'error compiling library: ', error);
  end;

var
  metalLibrary: TMetalLibrary;
  error       : NSError = nil;
  device      : MTLDeviceProtocol;
  extension   : string;
begin
  Fatal(CurrentThreadContext = nil, kError_InvalidContext);

  device:=CurrentThreadContext.device;
  metalLibrary:=TMetalLibrary.Create;
  extension:=ExtractFileExt(options.name);

  if options.code='' then begin
    extension:=ExtractFileExt(options.name);
    if extension='.metal' then metalLibrary.lib:=CompileShader(device, options.name)
    else if extension='.metallib' then metalLibrary.lib:=device.newLibraryWithFile_error(NSSTR(@options.name[1]), @error)
    else if options.name='' then metalLibrary.lib:=device.newDefaultLibrary
    else Fatal('invalid library name');

  end else begin
    metalLibrary.lib:=device.newLibraryWithSource_options_error(NSSTR(@options.code[1]), nil, @error);
    Fatal(metalLibrary.lib = nil, 'error compiling library: ', error);
  end;

  Fatal(metalLibrary.lib = nil, 'no metal shaders could be loaded.', error);
  Show(metalLibrary.lib);

  result:=metalLibrary;
end;

function LocalCreatePipeline(options: TMetalPipelineOptions): TMetalPipeline;
var
  libraryOptions    : TMetalLibraryOptions;
  colorAttachment   : MTLRenderPipelineColorAttachmentDescriptor;
  pipelineDescriptor: MTLRenderPipelineDescriptor;
  device            : MTLDeviceProtocol;
 // view              : MTKView;
  vertexFunction    : MTLFunctionProtocol = nil;
  fragmentFunction  : MTLFunctionProtocol = nil;
  kernelFunction    : MTLFunctionProtocol = nil;
  error             : NSError = nil;
begin
  Result:=TMetalPipeline.Create;
  device:=CurrentThreadContext.device;
 // view:=CurrentThreadContext.view;
  if options.shaderLibrary=nil then begin
    libraryOptions:=TMetalLibraryOptions.Create(options.libraryName, options.code);
    Result.shaderLibrary:=LocalCreateLibrary(libraryOptions);
  end else Result.shaderLibrary:=options.shaderLibrary;

  if options.kernelFunction<>'' then begin
    kernelFunction:=Result.shaderLibrary.GetFunction(options.kernelFunction);
    Fatal(kernelFunction = nil, 'kernel function not found.');

    Result.computePipelineState:=device.newComputePipelineStateWithFunction_error(kernelFunction, @error);
    Fatal(Result.computePipelineState = nil, 'pipeline creation failed.', error);
  end else begin
    vertexFunction:=Result.shaderLibrary.GetFunction(options.vertexShader);
    Fatal(vertexFunction = nil, 'vertex function not found.');

    fragmentFunction:=Result.shaderLibrary.GetFunction(options.fragmentShader);
    Fatal(fragmentFunction = nil, 'fragment function not found.');

    if options.pipelineDescriptor<>nil then pipelineDescriptor:=options.pipelineDescriptor
    else begin
      pipelineDescriptor:=MTLRenderPipelineDescriptor.alloc.init.autorelease;
      pipelineDescriptor.setDepthAttachmentPixelFormat({view.depthStencilPixelFormat}MTLPixelFormatDepth32Float);
      pipelineDescriptor.setVertexDescriptor(options.vertexDescriptor);
      pipelineDescriptor.colorAttachmentAtIndex(0).setPixelFormat({view.colorPixelFormat}MTLPixelFormatBGRA8Unorm);
    end;
    pipelineDescriptor.setVertexFunction(vertexFunction);
    pipelineDescriptor.setFragmentFunction(fragmentFunction);

    Result.renderPipelineState:=device.newRenderPipelineStateWithDescriptor_error(pipelineDescriptor, @error);
    Fatal(Result.renderPipelineState = nil, 'pipeline creation failed.', error);
  end;
end;

{ TMetalLibraryOptions }

constructor TMetalLibraryOptions.Create(_name,_code: string);
begin
  code:=_code;
  name:=_name;
end;

class function TMetalLibraryOptions.Default: TMetalLibraryOptions;
begin
  result.name:='';
  result.code:='';
end;

{ TMetalPipelineOptions }

class function TMetalPipelineOptions.Default: TMetalPipelineOptions;
begin
  result.libraryName:='';
  result.vertexShader:='vertexShader';
  result.fragmentShader:='fragmentShader';
  result.kernelFunction:='';
  result.vertexDescriptor:=nil;
  result.shaderLibrary:=nil;
  result.pipelineDescriptor:=nil;
end;

{ TMetalLibrary }

function TMetalLibrary.GetFunction (name: string): MTLFunctionProtocol;
var
  func: MTLFunctionProtocol;
begin
  if functions=nil then functions:=NSMutableDictionary.alloc.init;
  func:=functions.objectForKey(NSSTR(name));
  if func=nil then begin
    func:=lib.newFunctionWithName(NSSTR(name));
    if func<>nil then begin
      functions.setObject_forKey(func, NSSTR(name));
      func.release;
    end;
  end;
  result:=func;
end;

destructor TMetalLibrary.Destroy;
begin
  lib.release;
  functions.release;
  inherited;
end;

{ TMetalContext }

class function TMetalContext.SharedContext: TMetalContext;
begin
  result:=CurrentThreadContext;
end;

procedure TMetalContext.Draw;
begin
// view.draw;
end;

procedure TMetalContext.SetPreferredFrameRate(newValue: integer);
begin
  //view.setPreferredFramesPerSecond(newValue);
end;

procedure TMetalContext.SetColorPixelFormat(pixelFormat: MTLPixelFormat);
begin
  //view.setColorPixelFormat(pixelFormat);
end;

procedure TMetalContext.MakeCurrent;
begin
  CurrentThreadContext:=self;
end;

destructor TMetalContext.Destroy;
begin
  commandQueue.release;
  textureLoader.release;

  inherited;
end;   //}

{ TMetalPipeline }

destructor TMetalPipeline.Destroy;
begin
  renderPipelineState.release;
  computePipelineState.release;
  depthStencilState.release;
  if shaderLibrary<>nil then shaderLibrary.Free;
  inherited;
end;

{ TRTTexture }

constructor TRTTexture.Create(Width, Height: Integer);
var
  RTTexDesc: MTLTextureDescriptor;
begin
  w:=Width;
  h:=Height;

  //writeln('rtw, rth= ', w, ' ', h);

  RTTexDesc:=MTLTextureDescriptor.alloc.init.autorelease;
  RTTexDesc.setTextureType(MTLTextureType2D);
  RTTexDesc.setPixelFormat(MTLPixelFormatBGRA8Unorm);
  RtTexDesc.setMipMapLevelCount(9);
  {$ifdef UseManagedRT}
  RTTexDesc.setStorageMode(MTLStorageModeManaged);
  {$else}
  RTTexDesc.setStorageMode(MTLStorageModePrivate);
  {$endif}
  RTTexDesc.SetUsage(MTLTextureUsageShaderRead or MTLTextureUsageRenderTarget);

  RTTexDesc.setWidth(w);
  RTTexDesc.setHeight(h);
  RTTexDesc.setDepth(1);
  RT:=CurrentThreadContext.device.newTextureWithDescriptor(RTTexDesc);
  Fatal(RT=nil, 'RTtexture: newTextureWithDescriptor failed');

  RndPass:=nil;
  ColorPipeline:=nil;
end;

procedure TRTTexture.SaveToBmp(const FN: AnsiString);
var
  Buffer : Pointer;
  str    : TFileStream;
  pRegion: MTLRegion;
  Desc   : MTLTextureDescriptor;
  Copier : MTLBlitCommandEncoderProtocol;
  temp   : MTLTextureProtocol;
  Orig   : MTLOrigin;
  Sz     : MTLSize;
begin
  pRegion:=MTLRegionMake2D(0, 0, w, h);
  GetMem(Buffer, w*h*4);

  {$ifdef UseManagedRT}
//  CurrentThreadContext.commandBuffer:=CurrentThreadContext.commandQueue.commandBuffer;
  Copier:=CurrentThreadContext.commandBuffer.BlitCommandEncoder;  // ??
  Copier.synchronizeResource(RT);                                 // ??
  Copier.endEncoding;
  CurrentThreadContext.commandBuffer.commit;
  CurrentThreadContext.commandBuffer.WaitUntilCompleted;
  RT.getBytes_bytesPerRow_fromRegion_mipmapLevel(PInteger(Buffer), w*4, pRegion, 0);
  Copier:=nil;
  {$else}
//  CurrentThreadContext.commandBuffer:=CurrentThreadContext.commandQueue.commandBuffer;

  Desc:=MTLTextureDescriptor.alloc.init.autorelease;
  Desc.setTextureType(MTLTextureType2D);
  Desc.setPixelFormat(MTLPixelFormatBGRA8Unorm);
  Desc.setStorageMode(MTLStorageModeManaged);
  Desc.setWidth(w);
  Desc.setHeight(h);
  Desc.setDepth(1);
  temp:=CurrentThreadContext.device.newTextureWithDescriptor(Desc).autorelease;
  Fatal(temp=nil, 'GetData: Create Temp Texture failed');

  Copier:=CurrentThreadContext.commandBuffer.BlitCommandEncoder{.autorelease};
  Fatal(Copier=nil, 'GetData: create MTLBlitCommandEncoder failed');

  //Copier.synchronizeResource(RT);
  Orig.x:=0; Orig.y:=0; Orig.z:=0;
  Sz.width:=w; Sz.height:=h; Sz.depth:=1;
  Copier.copyFromTexture(RT, 0, 0, Orig, Sz, temp, 0, 0, Orig);
  //LogMetalError('Copy Form RenderTarget to Temp Texture');  b
  Copier.synchronizeResource(temp);

  Copier.endEncoding;
  CurrentThreadContext.commandBuffer.commit;
  CurrentThreadContext.commandBuffer.WaitUntilCompleted;
  CurrentThreadContext.commandBuffer:=nil;
  CurrentThreadContext.frameState:=[];

  writeln('w,h= ', w, ', ', h);
  temp.getBytes_bytesPerRow_fromRegion_mipmapLevel(PInteger(Buffer), w*4, pRegion, 0);
 // LogMetalError('Copy From Temp Texture to DIB');

 //  LogMetalError('Copy.EndEncoding');
  {$endif}

  //writeln('save bitmap');
  str:=TFileStream.Create(FN, fmCreate);
  str.size:=w*h*4+54;//54 bytes in header
//  str.Position:=0;
  str.WriteByte($42); str.WriteByte($4D); str.WriteByte($36); str.WriteByte($B0);
  str.WriteByte($04); str.WriteByte($00); str.WriteByte($00); str.WriteByte($00);
  str.WriteByte($00); str.WriteByte($00); str.WriteByte($36); str.WriteByte($00);
  str.WriteByte($00); str.WriteByte($00); str.WriteByte($28); str.WriteByte($00);

  str.WriteByte($00); str.WriteByte($00); str.WriteByte($40); str.WriteByte($01);
  str.WriteByte($00); str.WriteByte($00); str.WriteByte($F0); str.WriteByte($00);
  str.WriteByte($00); str.WriteByte($00); str.WriteByte($01); str.WriteByte($00);
  str.WriteByte($20); str.WriteByte($00); str.WriteByte($00); str.WriteByte($00);

  str.WriteByte($00); str.WriteByte($00); str.WriteByte($00); str.WriteByte($B0);
  str.WriteByte($04); str.WriteByte($00); str.WriteByte($00); str.WriteByte($00);
  str.WriteByte($00); str.WriteByte($00); str.WriteByte($00); str.WriteByte($00);
  str.WriteByte($00); str.WriteByte($00); str.WriteByte($00); str.WriteByte($00);

  str.WriteByte($00); str.WriteByte($00); str.WriteByte($00); str.WriteByte($00);
  str.WriteByte($00); str.WriteByte($00);
  str.Write(PByteArray(Buffer)^, w*h*4);

  FreeMem(buffer, w*h*4);
  str.Free;
end;

procedure TRTTexture.DrawTo({Tx: TGPUTexture; VTex: TVolumeTexture});
var
  verts       : array[0..3] of TAAPLVertex;
  viewportsize: array[0..1] of cuint;
 // VP          : MTLViewport;
  ZoomX, ZoomY, OffsetX, OffsetY, Zoom: single;
 // DiffColor   : TVec4;
begin
  InitShader;

  //begin scene
  CurrentThreadContext.commandBuffer:=CurrentThreadContext.commandQueue.commandBuffer;
  CurrentThreadContext.frameState+=[kMetalContextFrameStateOpen];
  LogMetalError('start1');
      //begin draw
      CurrentThreadContext.currentPipeline:=ColorPipeline;
      CurrentThreadContext.renderEncoder:=CurrentThreadContext.commandBuffer.renderCommandEncoderWithDescriptor(RndPass);

      LogMetalError('beginencoding1');
      CurrentThreadContext.renderEncoder.setCullMode(MTLCullModeNone);

      OffsetX:=0; OffsetY:=0; Zoom:=0.25;
      ZoomX:=w*Zoom;
      ZoomY:=h*Zoom;
      viewportsize[0]:=w;
      viewportsize[1]:=h;
      verts[0]:=AAPLVertex(V3(OffsetX-ZoomX, OffsetY+ZoomY,0), V4(0, 0, 0, 1));
      verts[1]:=AAPLVertex(V3(OffsetX-ZoomX, OffsetY-ZoomY,0), V4(0, 1, 0, 1));
      verts[2]:=AAPLVertex(V3(OffsetX+ZoomX, OffsetY+ZoomY,0), V4(1, 0, 0, 1));
      verts[3]:=AAPLVertex(V3(OffsetX+ZoomX, OffsetY-ZoomY,0), V4(1, 1, 0, 1));
      CurrentThreadContext.renderEncoder.setVertexBytes_length_atIndex(@verts, sizeof(verts), 0);
      CurrentThreadContext.renderEncoder.setVertexBytes_length_atIndex(@viewportSize, sizeof(viewportSize), 1);
      CurrentThreadContext.renderEncoder.setRenderPipelineState(ColorPipeline.renderPipelineState);
      CurrentThreadContext.renderEncoder.drawPrimitives_vertexStart_vertexCount(MTLPrimitiveTypeTriangleStrip, 0, 4);
      CurrentThreadContext.frameState+=[kMetalContextFrameStateRender];

    //end draw
    CurrentThreadContext.renderEncoder.endEncoding;
   // CurrentThreadContext.commandBuffer.commit;
    {$ifdef UseManagedRT}
    //CurrentThreadContext.commandBuffer.WaitUntilCompleted;
    {$endif}
   // CurrentThreadContext.commandbuffer:=nil;
    CurrentThreadContext.frameState:=[];
    //CurrentThreadContext.renderEncoder:=nil;
    LogMetalError('endencoding1');

  writeln('bundle path= ', bundleDir);

  SaveToBmp(BundleDir+'rt.bmp');
  //CurrentThreadContext.commandBuffer.commit;
 //! CurrentThreadContext.commandBuffer:=nil;
  CurrentThreadContext.frameState:=[];
  LogMetalError('present');
end;

procedure TRTTexture.InitShader;
var
  options    : TMetalPipelineOptions;
  RtDesc     : MTLRenderPipelineColorAttachmentDescriptor;
  colorAttach: MTLRenderPassColorAttachmentDescriptor;
begin
  if ColorPipeline=nil then begin
    options:=TMetalPipelineOptions.Default;
    options.code:=shaders.Color;

    options.pipelineDescriptor:=LocalCreatePipelineDescriptor;
    RtDesc:=options.pipelineDescriptor.colorAttachmentAtIndex(0);
    RtDesc.setBlendingEnabled(true);
   // RtDesc.setwriteMask(MTLColorWriteMaskGreen);
    RtDesc.setRgbBlendOperation(MTLBlendOperationAdd);
    RtDesc.setAlphaBlendOperation(MTLBlendOperationAdd);
    RtDesc.setSourceRGBBlendFactor(MTLBlendFactorSourceAlpha);
    RtDesc.setSourceAlphaBlendFactor(MTLBlendFactorSourceAlpha);
    RtDesc.setDestinationRGBBlendFactor(MTLBlendFactorOneMinusSourceAlpha);
    RtDesc.setDestinationAlphaBlendFactor(MTLBlendFactorOneMinusSourceAlpha);

    options.pipelineDescriptor.setSampleCount(4);
    ColorPipeline:=LocalCreatePipeline(options);
  end;

  //configure render target pass
  if RndPass=nil then begin
    RndPass:=MTLRenderPassDescriptor.alloc.init;
    colorAttach:=RndPass.colorAttachmentAtIndex(0);
    colorAttach.setTexture(RT);
    colorAttach.setClearColor(MTLClearColorMake(0.8, 0.7, 0.2, 1));
    colorAttach.setLoadAction(MTLLoadActionLoad);
    colorAttach.setStoreAction(MTLStoreActionStore);
  end;
end;

end.
